#include <stdlib.h>
#include "priority_queue.h"

#define PARENT_INDEX(x) ((x-1)>>1)
#define LEFT_INDEX(x) ((x<<1) + 1)
#define RIGHT_INDEX(x) ((x<<1) + 2)

void bubble_down(priority_queue *PQ, int j);
void bubble_up(priority_queue *PQ, int j);

inline void bubble_down(priority_queue *PQ, int j) {
    PQnode tmp;
    int i = j;

    // move an element initially at position j in the heap down in the
    // tree until the heap property is satisfied (read: as far down as
    // possible). The heap property is satisfied when the node currently
    // containing what was at j has no children, or its children are
    // "worse" than it.
    while (LEFT_INDEX(i) < PQ->size) {
        // current node has 2 children, right child is better than left child, and right better than current
        if (    RIGHT_INDEX(i) < PQ->size && 
                BETTER(heap[RIGHT_INDEX(i)], heap[LEFT_INDEX(i)]) && 
                BETTER(heap[RIGHT_INDEX(i)], heap[i])) 
        {
            SWAP(heap[RIGHT_INDEX(i)], heap[i], tmp);
            // follow the node, which is now where its right child was
            i = RIGHT_INDEX(i);
        // left better current and either left better than right or right does not exist
        } else if (BETTER(heap[LEFT_INDEX(i)], heap[i])) {
            SWAP(heap[LEFT_INDEX(i)], heap[i], tmp);
            // follow the node, which is now where its left child was
            i = LEFT_INDEX(i);
        } else {
            break;
        }
    }
}

inline void bubble_up(priority_queue *PQ, int j) {
    PQnode tmp;

    // swapping is needed if the node is better than the parent.
    for (   int i=j; 
            i > 0 && BETTER(heap[i], heap[PARENT_INDEX(i)]); 
            i = PARENT_INDEX(i)) {
        SWAP(heap[PARENT_INDEX(i)], heap[i], tmp);
        // keep looking at the same node, which is now where its parent was
    }
}

inline void pq_push(priority_queue *PQ, PQnode new_node) {
    int i = PQ->size;

#ifndef NO_RESIZE
    // ran out of space. Plz to be obtainz moar space.
    if (i == PQ->max_size) {
        PQ->max_size <<= 1; // double current capacity
        PQ->h = realloc(PQ->h, PQ->max_size*sizeof(PQnode *));
    }
#endif

    // place new node in first empty spot in the tree
    SET(heap[i], new_node);
    bubble_up(PQ, i);
    (PQ->size)++;
}

// assumes you will not try to empty an empty priority queue
PQnode pq_pop(priority_queue *PQ) {
    PQnode old_root;
    SET(old_root, heap[0]);

    // decrement size, then move "worst" node to the top and 
    // bubble it down to where it should belong.
    PQ->size -= 1;
    SET(heap[0], heap[PQ->size]);

    bubble_down(PQ, 0);

    return old_root;
}

priority_queue* pq_create(int initial_max_size) {
    priority_queue *PQ = malloc(sizeof(priority_queue));
    PQ->size = 0;
    PQ->max_size = initial_max_size;
    PQ->h = malloc(initial_max_size * sizeof(PQnode));
    return PQ;
}

priority_queue* pq_from_array(PQnode *array, int array_size) {
    priority_queue *PQ = malloc(sizeof(priority_queue));
    PQ->size = array_size;
    PQ->max_size = array_size;
    PQ->h = array;
    int i;

    // ensure the heap structure is preserved. We don't need to bubble down the second half
    // of the array, because these elements do not have children. Runs in O(n) time.
    for (i=(array_size-1)/2; i >= 0; i--) {
        bubble_down(PQ, i);
    }
    
    return PQ;
}


void pq_destroy(priority_queue *PQ) {
    int i;

    for (i=0; i < PQ->size; i++) {
        pqnode_destroy(heap[i]);
    }

    free(PQ->h);
    free(PQ);
}

#ifdef TEST
#include <stdio.h>
void pq_print(priority_queue *PQ) {
    int i;

    printf("Printing:\n");
    for (i=0; i < PQ->size; i++) {
        printf("%d: %f\n", i, HEUR(heap[i]));
    }
}

int main() {
    priority_queue *PQ = pq_create(8);
    PQnode nodes[4]; // declare nodes as array 3 of PQnode

#ifdef TEST_PTR
    for (int i=0; i<4; i++) {
        nodes[i] = malloc(sizeof(node));
    }
#endif

    HEUR(nodes[0]) = 0.0f;
    HEUR(nodes[1]) = 0.25f;
    HEUR(nodes[2]) = 0.5f;
    HEUR(nodes[3]) = 0.75f;

    pq_push(PQ, nodes[2] );
    pq_print(PQ);
    pq_push(PQ, nodes[0] );
    pq_print(PQ);
    pq_push(PQ, nodes[1] );
    pq_print(PQ);
    PQnode tmp = pq_pop(PQ);
    printf("%f\n", HEUR(tmp));
    pq_print(PQ);
    pq_pop(PQ);
    pq_print(PQ);
    pq_pop(PQ);
    pq_print(PQ);
    pq_push(PQ, nodes[3] );
    pq_print(PQ);
    //pq_destroy(PQ);
}
#endif
